# tag_release.py
import sys
import subprocess
import tomli # Or `tomli` for Python < 3.11

# 1. Read the version from the single source of truth
with open("pyproject.toml", "rb") as f:
    pyproject_data = tomli.load(f)
    version = pyproject_data["project"]["version"]

# 2. Construct the tag and the git command
tag = f"v{version}"
print(f"Found version {version}. Creating tag: {tag}")

# 3. Run the git command
try:
    subprocess.run(["git", "tag", tag], check=True)
except subprocess.CalledProcessError:
    print(f"Error: Could not create tag. Does the tag '{tag}' already exist?")
    sys.exit(1)
except FileNotFoundError:
    print("Error: 'git' command not found. Is Git installed and in your PATH?")
    sys.exit(1)
else:
    print(f"Successfully created tag '{tag}'.")
    sys.exit(0)


from pathlib import Path
from textual_slidecontainer import SlideContainer
from textual_slidecontainer.demo import SlideContainerDemo

DEMO_DIR = Path(__file__).parent.parent / "src" / "textual_slidecontainer"
TERINAL_SIZE = (110, 36)

async def test_launch():  
    """Test launching the SlideContainerDemo app."""
    app = SlideContainerDemo()
    async with app.run_test() as pilot:
        await pilot.pause()
        await pilot.exit(None) 

async def test_sliding_state():  
    """Test sliding containers open and close with key bindings."""
    app = SlideContainerDemo()
    async with app.run_test() as pilot:  

        assert app.query_one("#top_slidecontainer", SlideContainer).state is True
        await pilot.press("ctrl+w")  
        await pilot.wait_for_animation()
        assert app.query_one("#top_slidecontainer", SlideContainer).state is False

        assert app.query_one("#left_slidecontainer", SlideContainer).state is True
        await pilot.press("ctrl+a")  
        await pilot.wait_for_animation()
        assert app.query_one("#left_slidecontainer", SlideContainer).state is False

        assert app.query_one("#bottom_slidecontainer", SlideContainer).state is False
        await pilot.press("ctrl+s")  
        await pilot.wait_for_animation()
        assert app.query_one("#bottom_slidecontainer", SlideContainer).state is True

        assert app.query_one("#right_slidecontainer", SlideContainer).state is False
        await pilot.press("ctrl+d")  
        await pilot.wait_for_animation()
        assert app.query_one("#right_slidecontainer", SlideContainer).state is True    

def test_snapshot_only_left(snap_compare):

    async def run_before(pilot) -> None:
        await pilot.press("ctrl+a")  
        await pilot.wait_for_animation()

    assert snap_compare(
        DEMO_DIR / "demo.py",
        terminal_size=TERINAL_SIZE,
        run_before=run_before,
    )         
    
def test_snapshot_only_right(snap_compare):

    async def run_before(pilot) -> None:
        await pilot.press("ctrl+d")  
        await pilot.wait_for_animation()

    assert snap_compare(
        DEMO_DIR / "demo.py",
        terminal_size=TERINAL_SIZE,
        run_before=run_before,
    )         
    
def test_snapshot_left_and_right(snap_compare):

    async def run_before(pilot) -> None:
        await pilot.press("ctrl+a")  
        await pilot.press("ctrl+d")  
        await pilot.wait_for_animation()

    assert snap_compare(
        DEMO_DIR / "demo.py",
        terminal_size=TERINAL_SIZE,
        run_before=run_before,
    )         

def test_snapshot_top_and_bottom(snap_compare):

    async def run_before(pilot) -> None:
        await pilot.press("ctrl+w")  
        await pilot.press("ctrl+s")  
        await pilot.wait_for_animation()

    assert snap_compare(
        DEMO_DIR / "demo.py",
        terminal_size=TERINAL_SIZE,
        run_before=run_before,
    )             

def test_snapshot_all(snap_compare):

    async def run_before(pilot) -> None:
        await pilot.press("ctrl+w")  
        await pilot.press("ctrl+s")
        await pilot.press("ctrl+a")  
        await pilot.press("ctrl+d")          
        await pilot.wait_for_animation()

    assert snap_compare(
        DEMO_DIR / "demo.py",
        terminal_size=TERINAL_SIZE,
        run_before=run_before,
    )       


"""Contains the demo app."""

# ~ Type Checking (Pyright and MyPy) - Strict Mode
# ~ Linting - Ruff
# ~ Formatting - Black - max 110 characters / line

from __future__ import annotations

# from typing import cast

# Textual imports
from textual.app import App, ComposeResult
from textual import on
from textual.binding import Binding

from textual.widget import Widget
from textual.widgets import Header, Footer, Static, Button, RichLog, Switch
from textual.containers import Horizontal, Container

from textual_slidecontainer.slidecontainer import SlideContainer


class MySlideContainer(SlideContainer):

    def __init__(self) -> None:
        super().__init__(
            slide_direction="down",
            dock_position="bottomright",
            start_open=False,
            classes="slidecontainer bottom",
            id="bottom_slidecontainer",
        )

    def compose(self) -> ComposeResult:

        with Container():
            with Horizontal():
                yield Button("Hide", id="button_bottom")
                yield Switch(id="slide_dir_switch")
                yield Static(
                    "Change slide direction. \nCurrently: [yellow]down[/yellow]", id="slide_dir_label"
                )
            yield Static(
                "Fade is [red]off.[/red] " "Default is [yellow]closed.[/yellow] " "Menu is [yellow]floating."
            )

    @on(Switch.Changed)
    def handle_switch_change(self, event: Switch.Changed) -> None:
        """Change the slide direction based on the switch state."""
        if event.switch.value:
            self.set_slide_direction("right")
            self.query_one("#slide_dir_label", Static).update(
                "Change slide direction. \nCurrently: [yellow]right[/yellow]"
            )
        else:
            self.set_slide_direction("down")
            self.query_one("#slide_dir_label", Static).update(
                "Change slide direction. \nCurrently: [yellow]down[/yellow]"
            )


class SlideContainerDemo(App[None]):

    CSS_PATH = "demostyles.tcss"

    BINDINGS = [
        Binding("ctrl+w", "toggle_container('top')", "Top menu"),
        Binding("ctrl+a", "toggle_container('left')", "Left menu"),
        Binding("ctrl+s", "toggle_container('bottom')", "Bottom menu"),
        Binding("ctrl+d", "toggle_container('right')", "Right menu"),
        Binding("t", "toggle_borders", "Toggle main screen borders"),
    ]

    TITLE = "Textual-SlideContainer Demo"

    def __init__(self) -> None:
        super().__init__()
        self.main_container = Container(id="main_container")

    def compose(self) -> ComposeResult:

        yield Header()

        self.main_container.styles.opacity = 0.0  # the madlad loading screen
        with self.main_container:

            # There are 3 different ways to add children to a widget in Textual,
            # and all of them work with SlideContainer:
            # 1) Context manager
            # 2) Pass a list of widgets to the constructor
            # 3) Custom widget with compose method

            # 1) Context manager
            with SlideContainer(
                classes="slidecontainer top",
                id="top_slidecontainer",
                slide_direction="up",
                floating=False,  # Note this is True by default
                fade=True,  # and this is False by default
            ):
                yield Button("Hide", id="button_top")
                yield Static(
                    "Fade is [yellow]on.[/yellow] "
                    "Default is [yellow]open.[/yellow] "
                    "Menu is [red]not floating."
                )

            # 2) Create a list of widgets to pass to the next SlideContainer...
            window_widgets: list[Widget] = [
                Button("Hide", id="button_left"),
                Static(
                    "Fade is [red]off.[/red]\n"
                    "Default is [yellow]open.[/yellow]\n"
                    "Menu is [red]not floating."
                ),
            ]
            with Horizontal(id="horizontal_container"):

                # ...then pass a list of widgets to the constructor
                yield SlideContainer(
                    *window_widgets,
                    classes="slidecontainer leftright",
                    id="left_slidecontainer",
                    slide_direction="left",
                    floating=False,
                    duration=1.0,  # <-- you can change the animation duration.
                    easing_function="out_cubic",  # <-- you can change the easing function.
                )

                with Container(id="main_content"):
                    yield Static("This is content at the top left.", classes="top_docked_text")
                    yield Static(
                        "This is content \non the right \nthat can be blocked.", classes="right_docked_text"
                    )
                    yield RichLog()
                    yield Static(
                        "This is content at the bottom that can be blocked.",
                        classes="bottom_docked_text",
                    )

                with SlideContainer(
                    classes="slidecontainer leftright",
                    id="right_slidecontainer",  # Floating mode is the default.
                    slide_direction="right",  # When floating, It'll auto-dock to the same direction.
                    start_open=False,
                    fade=True,
                    dock_position="right",
                    # offset_y = 5      <-- you can enter a manual offset if you prefer.
                ):
                    yield Button("Hide", id="button_right")
                    yield Static(
                        "Fade is [yellow]on.[/yellow]\n"
                        "Default is [red]closed.[/red]\n"
                        "Menu is [yellow]floating."
                    )

            # 3) Custom widget with compose method
            yield MySlideContainer()

        yield Footer()

    def on_mount(self) -> None:
        """Called when the app is mounted. This is a good place to set up event listeners."""
        self.query_one(RichLog).can_focus = False
        self.call_after_refresh(self.finished_loading)

    def action_toggle_container(self, direction: str) -> None:
        """Toggle the slidecontainer open and closed."""
        slidecontainer = self.query_one(f"#{direction}_slidecontainer", SlideContainer)
        slidecontainer.toggle()

    def action_toggle_borders(self) -> None:
        """Toggle the borders of the slidecontainers."""
        # self.query_one("#main_container", Container).toggle_class("bordered")
        self.query_one("#horizontal_container", Horizontal).toggle_class("bordered")
        self.query_one("#main_content", Container).toggle_class("bordered")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Toggle the slidecontainer open and closed."""
        assert event.button.id  # for type checking
        direction = event.button.id.split("_")[1]  # remove the `button_` prefix
        self.action_toggle_container(direction)

    def finished_loading(self) -> None:
        """This is a madlad way of making a loading screen. The main container starts
        at opacity 0.0 and fades in to 1.0 when the slidecontainer is done loading."""

        self.main_container.styles.animate("opacity", value=1.0, duration=0.3)
        # self.main_container.styles.opacity = 1.0     # this would be the simpler way of doing it.

    @on(SlideContainer.SlideCompleted)
    def slide_completed(self, event: SlideContainer.SlideCompleted) -> None:

        rich_log = self.query_one(RichLog)
        rich_log.write(f"{event.container.id} {'opened' if event.state else 'closed'}")


def run_demo() -> None:
    SlideContainerDemo().run()


if __name__ == "__main__":
    run_demo()



"Module for the SlideContainer widget for Textual."

# ~ Type Checking (Pyright and MyPy) - Strict Mode
# ~ Linting - Ruff
# ~ Formatting - Black - max 110 characters / line

from __future__ import annotations
from typing import Literal, get_args, cast
import asyncio

# Textual imports
from textual import work
from textual.containers import Container
from textual.geometry import Offset, Size
from textual.reactive import reactive
from textual.message import Message
from textual.widget import Widget
from textual._animator import Animatable
import textual.events as events

SLIDE_DIRECTION = Literal["left", "right", "up", "down"]
DOCK_POSITION = Literal[
    "topleft",
    "top",
    "topright",
    "left",
    "right",
    "bottomleft",
    "bottom",
    "bottomright",
    "none",
]

EASING_FUNC = Literal[
    "none",
    "round",
    "linear",
    "in_sine",
    "in_out_sine",
    "out_sine",
    "in_quad",
    "in_out_quad",
    "out_quad",
    "in_cubic",
    "in_out_cubic",
    "out_cubic",
    "in_quart",
    "in_out_quart",
    "out_quart",
    "in_quint",
    "in_out_quint",
    "out_quint",
    "in_expo",
    "in_out_expo",
    "out_expo",
    "in_circ",
    "in_out_circ",
    "out_circ",
    "in_back",
    "in_out_back",
    "out_back",
    "in_elastic",
    "in_out_elastic",
    "out_elastic",
    "in_bounce",
    "in_out_bounce",
    "out_bounce",
]


class SlideContainer(Container):
    """See init for usage and information."""

    class InitCompleted(Message):
        """Message sent when the container is ready.
        This is only sent if the container is starting closed."""

        def __init__(self, container: SlideContainer) -> None:
            super().__init__()
            self.container = container
            """The container that is ready."""

        @property
        def control(self) -> SlideContainer:
            """The SlideContainer that sent the message."""
            return self.container

    class SlideCompleted(Message):
        """Message sent when the container is opened or closed.
        This is sent after the animation is complete."""

        def __init__(self, state: bool, container: SlideContainer) -> None:
            super().__init__()
            self.state = state
            """The state of the container.  \n True = container open, False = container closed."""
            self.container = container
            """The container that has finished sliding."""

        @property
        def control(self) -> SlideContainer:
            """The SlideContainer that sent the message."""
            return self.container

    state: reactive[bool] = reactive[bool](True)
    """State of the container.  \n True = container open, False = container closed.   
    You can set this directly, or you can use the toggle() method."""

    _current_layer: int = 0

    def __init__(
        self,
        *children: Widget,
        slide_direction: SLIDE_DIRECTION,
        dock_position: DOCK_POSITION = "none",
        floating: bool = True,
        start_open: bool = True,
        fade: bool = False,
        offset_x: int = 0,
        offset_y: int = 0,
        duration: float = 0.8,
        easing_function: EASING_FUNC = "out_cubic",
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ):
        """Construct a Sliding Container widget.

        Args:
            *children: Child widgets.
            slide_direction: Can be:
                - left
                - right
                - up
                - down
                NOTE: This is not tied to dock position. Feel free to experiment.
            dock_position: The position to dock the container. Can be:
                - topleft
                - top
                - topright
                - left
                - right
                - bottomleft
                - bottom
                - bottomright
                - none
                NOTE: When floating is True, this is automatically set to the same direction
                as the slide direction. (up = top, down = bottom, left = left, right = right)
                Floating SlideContainers MUST be docked to a direction. However, you can change the dock direction.
                The dock direction does not need to be the same as the slide direction.
            floating: Whether the container should float overtop on its own layer.
            start_open: Whether the container should start open(visible) or closed(hidden).
            fade: Whether to also fade the container when it slides.
            offset_x: The
            offset_y: The
            duration: The duration of the slide animation in seconds.
            easing_function: The easing function to use for the animation.
            name: The name of the widget.
            id: The ID of the widget in the DOM.
            classes: The CSS classes for the widget.
            disabled: Whether the widget is disabled or not.
        """
        super().__init__(*children, name=name, id=id, classes=classes, disabled=disabled)

        if slide_direction not in get_args(SLIDE_DIRECTION):
            raise ValueError("slide_direction must be one of 'left', 'right', 'up', or 'down'.")
        # if dock_direction not in get_args(DOCK_DIRECTION):
        #     raise ValueError("dock_direction must be one of 'left', 'right', 'top', 'bottom', or 'none'.")
        if dock_position not in get_args(DOCK_POSITION):
            raise ValueError(
                "dock_position must be one of 'topleft', 'top', 'topright', "
                "'left', 'right', 'bottomleft', 'bottom', 'bottomright', or 'none'."
            )

        if easing_function not in get_args(EASING_FUNC):
            raise ValueError(
                "easing_function must be one of the allowed functions. See the Textual docs for more info."
            )
        if slide_direction in ["left", "right"] and offset_x != 0:
            raise ValueError(
                "When slide_direction is 'left' or 'right', offset_x must be 0. "
                "Use offset_y to adjust the vertical position."
            )
        if slide_direction in ["up", "down"] and offset_y != 0:
            raise ValueError(
                "When slide_direction is 'up' or 'down', offset_y must be 0. "
                "Use offset_x to adjust the horizontal position."
            )

        self.slide_direction = str(slide_direction)
        self.floating = floating
        self.set_reactive(SlideContainer.state, start_open)  # need to handle this manually
        self.fade = fade
        self.dock_position = str(dock_position)  # default is "none" - but only if floating is False.
        self.offset_x = offset_x
        self.offset_y = offset_y
        self.offset = Offset(offset_x, offset_y)  # set the initial offset
        self.duration = duration
        self.easing_function = easing_function

        self.layer_index = SlideContainer._current_layer
        SlideContainer._current_layer += 1  # increment the class variable for the next window's layer

        if self.floating:
            self.styles.layer = f"sliding_container{self.layer_index}"

            slide_dir_to_dock_pos: dict[str, str] = {
                "up": "top",
                "down": "bottom",
                "left": "left",
                "right": "right",
            }
            if dock_position == "none":  # NOTE: If floating, then it must be docked *somewhere*.
                self.dock_position = slide_dir_to_dock_pos[slide_direction]

        dock_pos_to_style_dock = {
            "topleft": "top",
            "top": "top",
            "topright": "right",  # note topright = right
            "left": "left",
            "right": "right",
            "bottomleft": "bottom",
            "bottom": "bottom",
            "bottomright": "right" if slide_direction in ["left", "right"] else "bottom",
            "none": "none",
        }

        # if starting closed, do a little visual trickery:
        if start_open is False:
            self.styles.opacity = 0.0

        self.styles.dock = dock_pos_to_style_dock[self.dock_position]

    ##################
    # ~ Public API ~ #
    ##################

    def open(self) -> None:
        "Open the container. This is the same as setting state to True."
        self.state = True

    def close(self) -> None:
        "Close the container. This is the same as setting state to False."
        self.state = False

    def toggle(self) -> None:
        "Toggle the state of the container. Opens or closes the container."
        self.state = not self.state

    def set_slide_direction(self, direction: str) -> None:
        """Set the slide direction of the container.

        Args:
            direction: The new slide direction. Must be one of 'left', 'right', 'up', or 'down'.
        """
        if direction not in get_args(SLIDE_DIRECTION):
            raise ValueError("slide_direction must be one of 'left', 'right', 'up', or 'down'.")
        self.slide_direction = direction
        if self.dock_position == "bottomright":
            current_dock = self.styles.dock
            if direction in ["up", "down"]:
                self.styles.dock = "bottom" if direction == "down" else "top"
            else:
                self.styles.dock = "right" if direction == "right" else "left"
            if self.styles.dock != current_dock:
                self.set_to_starting_position()

    #################
    # ~ Internals ~ #
    #################

    def _on_mount(self, event: events.Mount) -> None:
        super()._on_mount(event)
        self.set_to_starting_position()

    @work
    async def set_to_starting_position(self) -> None:

        offset_x, offset_y = await self._calculate_offsets()

        if self.state is False:  # This means the container is starting closed.

            if self.slide_direction == "left":
                self.styles.offset = Offset(-(self.size.width + self.get_spacing("horizontal")), offset_y)
            elif self.slide_direction == "right":
                self.styles.offset = Offset(self.size.width + self.get_spacing("horizontal"), offset_y)
            elif self.slide_direction == "up":
                self.styles.offset = Offset(offset_x, -(self.size.height + self.get_spacing("vertical")))
            elif self.slide_direction == "down":
                self.styles.offset = Offset(offset_x, self.size.height + self.get_spacing("vertical"))

            self.display = False
            self.styles.opacity = 1.0  #  Was set to 0 earlier. Must change back.

        else:  # This means the container is starting open.
            self.styles.offset = Offset(offset_x, offset_y)

        self.post_message(self.InitCompleted(self))  # Notify that the container is ready.

    async def watch_state(self, old_state: bool, new_state: bool) -> None:

        if new_state == old_state:
            return
        if new_state is True:
            await self._slide_open()
        else:
            await self._slide_closed()

    def get_spacing(self, direction: str) -> int:

        if direction == "horizontal":
            return self.styles.border.spacing.left + self.styles.border.spacing.right
        elif direction == "vertical":
            return self.styles.border.spacing.top + self.styles.border.spacing.bottom
        else:
            raise ValueError("Invalid direction for get_spacing func. Must be 'horizontal' or 'vertical'.")

    def _slide_open_completed(self) -> None:
        self.post_message(self.SlideCompleted(True, self))

    def _slide_closed_completed(self) -> None:
        self.display = False
        self.post_message(self.SlideCompleted(False, self))

    async def _slide_open(self) -> None:

        # This is here just in case anyone calls this method manually:
        if self.state is not True:
            self.set_reactive(SlideContainer.state, True)  # set state without calling the watcher

        self.display = True
        offset_x, offset_y = await self._calculate_offsets()

        self.animate(
            "offset",
            cast(Animatable, Offset(offset_x, offset_y)),
            duration=self.duration,
            easing=self.easing_function,
            on_complete=self._slide_open_completed,
        )
        if self.fade:
            self.styles.animate(
                "opacity", value=1.0, duration=self.duration, easing=self.easing_function
            )  # reset to original opacity

    async def _slide_closed(self) -> None:

        # This is here just in case anyone calls this method manually:
        if self.state is not False:
            self.set_reactive(SlideContainer.state, False)  # set state without calling the watcher

        offset_x, offset_y = await self._calculate_offsets()

        # NOTE: The casting to Animatable here is necessary because of an apparent mismatch
        # in type hiting between python versions. When type checking this file using python 3.9,
        # MyPy complains about Offset and says it is not an Animatable type.
        # This problem does not happen in python 

        if self.slide_direction == "left":
            self.animate(
                "offset",
                cast(Animatable, Offset(-(self.size.width + self.get_spacing("horizontal")), offset_y)),
                duration=self.duration,
                easing=self.easing_function,
                on_complete=self._slide_closed_completed,
            )
        elif self.slide_direction == "right":
            self.animate(
                "offset",
                cast(Animatable, Offset(self.size.width + self.get_spacing("horizontal"), offset_y)),
                duration=self.duration,
                easing=self.easing_function,
                on_complete=self._slide_closed_completed,
            )
        elif self.slide_direction == "up":
            self.animate(
                "offset",
                cast(Animatable, Offset(offset_x, -(self.size.height + self.get_spacing("vertical")))),
                duration=self.duration,
                easing=self.easing_function,
                on_complete=self._slide_closed_completed,
            )
        elif self.slide_direction == "down":
            self.animate(
                "offset",
                cast(Animatable, Offset(offset_x, self.size.height + self.get_spacing("vertical"))),
                duration=self.duration,
                easing=self.easing_function,
                on_complete=self._slide_closed_completed,
            )

        # NOTE: The offsets use self.animate, while the opacity uses self.styles.animate.
        # The reason is because self.offset is a setter method on the Widget class that provides
        # a shortcut to the styles.offset property and generates an Offset object from a tuple.
        # Trying to use styles.animate("offset", ...) would not work. I'm not entirely sure why,
        # but simply using the setter method on the widget does work. Presumably that's why it
        # was even added to the Widget class in the first place.

        if self.fade:
            self.styles.animate("opacity", value=0.0, duration=self.duration, easing=self.easing_function)

    async def _calculate_offsets(self) -> tuple[int, int]:
        """Returns the starting position of the window
        based on the parent size and starting position arguments."""

        # This check is to ensure that the size is set before calculating offsets.
        # It fixes a bug caused by setting display to false and then setting it back to true
        # a split second before the offsets are calculated.
        if self.size == Size(0, 0):
            await asyncio.sleep(0.01)
            await self._calculate_offsets()

        assert isinstance(self.parent, Widget)
        max_x, max_y = self.parent.size - self.size
        max_x = max_x - self.get_spacing("horizontal")
        max_y = max_y - self.get_spacing("vertical")

        if self.dock_position in ["left", "right"]:
            horizontal = 0
            vertical = max_y // 2
        elif self.dock_position in ["top", "bottom"]:
            horizontal = max_x // 2
            vertical = 0
        elif self.dock_position == "bottomright":
            if self.slide_direction in ["left", "right"]:  # will be docked to the right
                horizontal = 0
                vertical = max_y
            else:  # slide_direction in ["up", "down"],  will be docked to the bottom
                horizontal = max_x
                vertical = 0
        else:
            horizontal = 0
            vertical = 0

        x_to_return = self.offset_x if self.offset_x else horizontal
        y_to_return = self.offset_y if self.offset_y else vertical
        return x_to_return, y_to_return



"""Textual-SlideContainer
Copyright (c) 2025 Edward Jazzhands.
Licensed under the MIT License.

To use:
```
from textual_slidecontainer import SlideContainer
```

Package standards:
- Type Checking (Pyright and MyPy) - strict
- Linting - Ruff
- Formatting - Black - max 110 characters / line
"""

from importlib.metadata import version as get_version
from textual_slidecontainer.slidecontainer import SlideContainer

__all__ = [
    "SlideContainer",
]
__version__ = get_version("textual_slidecontainer")



"""Config file for Nox sessions
By Edward Jazzhands - 2025

NOTE ABOUT NOX CONFIG:
UV and Nox work amazing together. But in some cases, such as if you are
doing development inside a docker container or in other such niche environments,
UV will not be able to create its normal symlinks when making .venv folders.
(For example you might bind mount a folder into the container which contains
your project, you cannot symlink between the container's filesystem and this folder.
That's a fundamental Linux limitation, not because of UV or Docker.)

This creates a problem where Nox will create a new venv folder with hard
copies of everything for every single test. Nox by default does not reuse
existing virtual environments, but rather it will create a new one for each
session every single time it runs. If you have the aforementioned situation
with UV and the symlinks, this means that Nox will be writing several GB
of data to your drive every single time you run it. The space usage itself is 
not so much an issue as the large amount of writing wear it will cause on your
drive if you are writing anywhere from 2-10gb of data every time you run nox.

Setting `nox.options.reuse_existing_virtualenvs` helps to mitigate this issue.
Nox will reuse environments between runs, preventing however
many GB of data from being written to your drive every time you run nox.

If you do not need to reuse existing virtual environments, you can set
`nox.options.reuse_existing_virtualenvs = False` and `DELETE_VENV_ON_EXIT = True`
to delete the virtual environments after each session. This will ensure that
you do not have any leftover virtual environments taking up space on your drive.
Nox would just delete them when starting a new session anyway.
"""

import nox
import pathlib
import shutil

# PYTHON_VERSIONS = ["3.9", "3.10", "3.11", "3.12"]
PYTHON_VERSIONS = ["3.9"]
MAJOR_TEXTUAL_VERSIONS = [3, 4, 5]

##############
# NOX CONFIG #
##############

nox.options.reuse_existing_virtualenvs = True
nox.options.stop_on_first_error = True
DELETE_VENV_ON_EXIT = False

if nox.options.reuse_existing_virtualenvs and DELETE_VENV_ON_EXIT:
    raise ValueError(
        "You cannot set both `nox.options.reuse_existing_virtualenvs`"
        "and `DELETE_VENV_ON_EXIT` to True (Technically this would not cause "
        "an error, but it would be pointless)."
    )

################
# NOX SESSIONS #
################

@nox.session(
    venv_backend="uv",
    python=PYTHON_VERSIONS,
)
@nox.parametrize("textual", MAJOR_TEXTUAL_VERSIONS)
def tests(session: nox.Session, textual: int) -> None:

    session.run_install(
        "uv",
        "sync",
        "--quiet",
        "--reinstall",
        f"--python={session.virtualenv.location}",
        env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location},
        external=True,
    )

    session.run_install(
        "uv", "pip", "install",
        f"textual<{str(textual + 1)}.0.0",
        external=True,
    )
    # EXPLANATION: The `textual + 1` is a trick to make UV
    # only download the last version of each major revision of Textual.
    # If the current version is 3, we're saying `install textual<4.0.0`.
    # This will make UV grab the highest version of Textual 3.x.x, which is 3.7.1.
    
    session.run("uv", "pip", "show", "textual")

    session.run("ruff", "check", "src")
    session.run("mypy", "src")
    session.run("basedpyright", "src")
    session.run("pytest", "tests", "-vvv") 

    session_path = pathlib.Path(session.virtualenv.location)
    if session_path.exists() and session_path.is_dir() and DELETE_VENV_ON_EXIT:
        shutil.rmtree(session_path)
